<style>
    /* Sposta i pulsanti a lato (a sinistra) */
    #buttons {
        position: absolute;
        left: 10px;
        top: 50px;
        z-index: 999;
    }

    /* Ridimensiona il grafo per occupare il resto della finestra */
    #mynetwork {
        position: absolute;
        left: 200px;
        top: 0;
        right: 0;
        bottom: 0;
        border: none;  /* Rimuove il bordo */
        width: calc(100% - 200px); /* Occupa tutta la larghezza */
        height: 100vh; /* Occupa tutta l'altezza */
    }

    /* Rimuovi il bordo della configurazione */
    #config {
        border: none; /* Rimuove il bordo del contenitore dei pulsanti */
    }
</style>